/*
   program : SheppLogan

   $Source: /home/yee/C/RCS/SheppLogan.c,v $
   $Author: yee $
   $Date: 1999/02/10 16:39:47 $
   $Revision: 2.26 $

   This program implements the Shepp-Logan head phantom as defined in
   Kak & Slaney, "Principles of Computerized Tomographic Imaging",
   pgs. 53-6.

   Note : the distance between elements of a projection is to be stored
          in h->tau[X].
   Note : the angle between projections is to be stored in h->tau[Y].

   Default output is stdout.

   Parameters : see documention for ParseArg()

   Assumptions : none known

   Bugs : none known
     
   To Do :
     1) Speed optimizations needed.  Program now calls trig functions
        more often than needed.

   $Log: SheppLogan.c,v $
   Revision 2.26  1999/02/10 16:39:47  yee
   *** empty log message ***

   Revision 2.25  1999/02/06 20:03:07  yee
   bug fix in restricting # elements to power of 2.

   Revision 2.24  1999/02/05 04:08:34  yee
   *** empty log message ***

   Revision 2.23  1999/01/23 04:05:29  yee
   restrict elements/projection to power of 2.


   Revision 2.2  1997/06/21 02:36:27  yee
   bug fix in simulating parallel beam data as fan beam.

   Copyright 1997,1998,1999 by Conway Yee.  All rights reserved.
 */

char ShepLogan[] = "$Id: SheppLogan.c,v 2.26 1999/02/10 16:39:47 yee Exp $";

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <math.h>

#include "mutil.h"
#include "util.h"
#include "array.h"
#include "style.h"
#include "codes.h"

#include "recon.h"
#include "ImageHeader.h"
#include "SheppLogan.h"

PUBLIC char *optarg;               /* used by getopt() */
PUBLIC int optind=0;
PUBLIC int opterr=0;
PUBLIC int optopt=0;

PRIVATE ImageHeader *ParseArg(ImageHeader *,int,char **,FILE **);
PRIVATE void  OutputProjectionData(FILE *,ImageHeader *);
PRIVATE RECON_FLOAT proj(double,double);

BOOLEAN verbose=FALSE;  /* print out routine messages? */
 
/*****************************************************************************/
/*****************************************************************************/

PUBLIC int main(int argc,char **argv)

{ /* main() */
  FILE *outptr=stdout; /* output file pointer. use stdout as default */
  ImageHeader *h;      /* header to projection data */

  SetupSL();  /* set up Shepp-Logan phantom */

  AllocateImageHeader(h,NDIM);
  ImageHeaderEditPersonalData(h->name,"Shepp-Logan Head Phantom");
  ImageHeaderEditPersonalData(h->loc,"IEEE Transactions Nuclear Science");
  ImageHeaderEditPersonalData(h->mr,"NS-21, 6/74, pgs 21-43");

  h=ParseArg(h,argc,argv,&outptr);

  if(!ioImageHeader(h,fwrite,outptr)) {
    errno=HEADER_ERROR;
    perror("SheppLogan : error in output of header ");
    return(errno);
  }

  OutputProjectionData(outptr,h);

  FreeImageHeader(h);
  fclose(outptr);

  return(errno);
} /* main() */

/*****************************************************************************/

/* this function generates the projection data for the appropriate CT mode

   NOTE: theta is of the angle between the line o projection and the
         with respect to +X axis (not the ray perpendicular to it).

   Note: central angle of projection is 90 degrees off from theta, the angle
         of projection. */

PRIVATE void OutputProjectionData
(
 FILE *outptr,      /* output file pointer */
 ImageHeader *h     /* reconstruction header */
)
{
  RECON_FLOAT p;                    /* projection at line theta,r */

                                /* defined to avoid array accesses */
  int hi_ele=h->lim[HI_ELE];        /* max index of projection element */
  int hi_proj=h->lim[HI_PROJ];      /* max index of projection count */

  switch(h->type) {

#ifdef RECON_PARALLEL
  case PARALLEL:
    { /* parallel data */
      int i,j;                     /* counter */
      double tau_X=h->tau[ELE];    /* distance between elements */
      double tau_Y=h->tau[PROJ];   /* distance between projections */

                            /* outer loop is each projection.
                               inner loop is each element in that projection */
      for(j=h->lim[LO_PROJ];j<=hi_proj;j++) { /* each projection */
        double theta=tau_Y*j;  /* projection plane angle */

        for(i=h->lim[LO_ELE];i<=hi_ele;i++) { /* elements in each projection */
          double r=tau_X*i;    /* projection offset */

          p=proj(theta-PI_2,r);
          fwrite(&p,sizeof(RECON_FLOAT),1,outptr);
        } /* elements in each projection */

      } /* each projection */
    } /* parallel data */
    break;
#endif RECON_PARALLEL

#ifdef RECON_ANGLE_FAN
  case ANGLE:
    { /* equiangular fan data */
      int i,j; /* counter */
      double gammad=h->tau[ELE];    /* angle between elements in projection */
      double betad=h->tau[PROJ];    /* angle between projections */
      double D=h->D;                /* distance between focal point and
                                       center of rotation */

                            /* outer loop is each projection.
                               inner loop is each element in that projection */
      for(j=h->lim[LO_PROJ];j<=hi_proj;j++) { /* each projection */
        double beta=j*betad-PI_2;    /* central angle of projection */

        for(i=h->lim[LO_ELE];i<=hi_ele;i++) { /* elements in each projection */
          double gamma=i*gammad;    /* angle of ray wrt central angle */
          double theta=beta+gamma;  /* projection angle */
          double r=D*sin(gamma);    /* projection offset */

          p=proj(theta,r);
          fwrite(&p,sizeof(RECON_FLOAT),1,outptr);
        } /* elements in each projection */

      } /* each projection */
    } /* equiangular fan data */
    break;
#endif RECON_ANGLE_FAN

#ifdef RECON_COLIN_FAN
  case COLIN:
    { /* colinear fan data */
      int i,j; /* counter */
      double spacing=h->tau[ELE];     /* distance between elements along
                                         detector array centered at origin */
      double betad=h->tau[PROJ];      /* angle between projections */
      double D=h->D;                  /* distance between focal point and
                                         center of rotation */

                            /* outer loop is each projection.
                               inner loop is each element in that projection */
      for(j=h->lim[LO_PROJ];j<=hi_proj;j++) { /* each projection */
        double beta=j*betad-PI_2;     /* central angle of projection */

        for(i=h->lim[LO_ELE];i<=hi_ele;i++) { /* elements in each projection */
          double s = i*spacing;          /* angle of ray */

          double r = s*D/SQRT(D*D+s*s);  /* distance of ray from origin */
          double sin_g = r/D;            /* sin(gamma) */
          double gamma = asin(sin_g);    /* angle of ray wrt central ray */
          double theta=beta+gamma;       /* projection angle */

          p=proj(theta,r);
          fwrite(&p,sizeof(RECON_FLOAT),1,outptr);
        } /* elements in each projection */

      } /* each projection */
    } /* colinear fan data */
    break;
#endif RECON_COLIN_FAN

  default:
    errno = UNSUPPORTED_MODE;
    perror("SheppLogan : unsupported mode");
    break;
  }
}

/*****************************************************************************/

/* this function returns the projection the Shepp-Logan head phantom onto
   a line of angle theta (wrt +X axis), and distance ro (polar coordinates).

   With the exception of theta and ro, the notation is from Kak & Slaney
   pgs 53-6.  */

PRIVATE RECON_FLOAT proj
(
 double theta,        /* projection angle (polar coordinates) */
 double ro            /* projection distance (polar coordinates) */
)
{ /* proj() */
  double p=0.0;       /* projection along the ray */

  { /* kludge */
    /*    theta *= -1;
          ro *= -1;*/
  } /* kludge */

  { /* cycle through each ellipse */
    int el;           /* ellipse counter */

    for(el=0;el<MAX_SL;el++) {
      double a2;      /* a^2(theta) Kak & Slaney p. 56 (eqn 5) */
      double ct;      /* t corrected for ellipse rotation */
      double ctheta;  /* theta corrected for ellipse rotation */

      ctheta=theta-ALPHA;             /* Kak & Slaney p. 56 (eqn 6) */

      ct=ro-S*cos(ATAN_YX-theta);     /* Kak & Slaney p. 56 (eqn 6) */
      ct *= ct;                       /* only t^2 used in equation */

      a2 = A*A*pow(cos(ctheta),2.0)+B*B*pow(sin(ctheta),2.0);
      if(a2>ct)
        p += 2.0*RO*A*B/a2*pow(a2-ct,0.5);  /* Note: ct already squared */
    } 
  } /* cycle through each ellipse */

  return(p);
} /* proj() */

/*****************************************************************************/

/* This function processes the command line arguments of SheppLogan.
   See the declaration of *errormsg[] below for legal options. */

PRIVATE ImageHeader *ParseArg
(
 ImageHeader *h,     /* image header */
 int argc,           /* argument count */
 char **argv,        /* arguments */
 FILE **outptr       /* output file pointer */
)
{ /* ParseArg() */
  char opt;                                        /* option */
  double AngleTraversed = DEFAULT_ANGLE_TRAVERSED; /* project thru PI vs 2PI */
  char *fname=NULL;                                /* output file name */
  BOOLEAN error=FALSE;                        /* error in parsing arguments */
  BOOLEAN ParallelSim=FALSE; /* simulate parallel beam as limit of fan beam? */
  double fov = 0.0;                           /* field of view */

  char SheppLoganOptions[]="e:p:f:d:t:o:c:v";
  char *errormsg[]={
    "SheppLogan [options]",
    "   -e [number of elements in a projection]",
    "   -p [number of projections]",
    "   -f [field of view]",
    "   -d [distance between focal point and center of rotation]",
    "   -t PARALLEL | ANGLE | COLIN  or CT mode number",
    "       ",
    "   -o [output filename]",
    "   -c fraction of a circle that is projected",
    "   -v verbose mode",
    "",
    "  Note: if two types given and one is PARALLEL, then data is simulated",
    "        as limit of other type",
    "",
    "  Note: the number of elements per projection is restricted to a",
    "        power of two",
    NULL
  };

  do {
    opt = getopt(argc,argv,SheppLoganOptions);
    switch(opt) {
    case 'e':                       /* number of elements in each projection */
      {
        int num;                     /* number read from command line option */

        num = atoi(optarg);

        {                              /* restrict # elements to  power of 2 */
          int size;  /* dummy variable */

          for(size=1;size<num;size<<=1);
          num=size;
        }

        h->dim[ELE] = num;
        h->lim[LO_ELE] = 1 - num/2;
        h->lim[HI_ELE] = num/2;
        if ((errno == ERANGE)||(h->lim[HI_ELE]<0)) {
          perror("SheppLogan : ParseArg : illegal elements # in projection ");
          error = TRUE;
        }
      }
      break;
    case 'p':                                       /* number of projections */
      h->dim[PROJ] = atoi(optarg);
      h->lim[LO_PROJ] = 0;
      h->lim[HI_PROJ] = h->dim[PROJ]-1;
      if ((errno == ERANGE)||(h->lim[HI_PROJ]<0)) {
        perror("SheppLogan : ParseArg : illegal number of projections ");
        error = TRUE;
      }
      break;
    case 'f':                                               /* field of view */
      fov = atof(optarg);
      if ((errno == ERANGE)||(fov<0)) {
        perror("SheppLogan : ParseArg : illegal field of view ");
        error = TRUE;
      }
      break;
    case 'd':  /* distance between focal point and rotation angle */
      h->D = atof(optarg);
      if ((errno == ERANGE)||(h->D<0)) {
        perror("SheppLogan : ParseArg : illegal focal point to rotation angle distance ");
        error = TRUE;
      }
      break;
    case 't':  /* CT mode */
      {
        CHOICE OldType = h->type;

        h->type = atoi(optarg);  /* assume numerical CT type */

                                                   /* assume text CT type */
        if((h->type <= IMAGE_TYPE_NOT_SET) || (h->type >= MAX_IMAGE_TYPE)) {
          STOUPPER(optarg);
          if (!strcmp(optarg,"PARALLEL"))
            h->type = PARALLEL;
          else if (!strcmp(optarg,"ANGLE"))
            h->type = ANGLE;
          else if (!strcmp(optarg,"COLIN"))
            h->type = COLIN;
          else {
            perror("ParseArg : illegal CT mode");
            error = TRUE;
          }
        }
                   /* simulate parallel beam data as limit of fanbeam data */
        if((OldType) && (OldType!=h->type) &&
           ((OldType==PARALLEL)||(h->type==PARALLEL)) ) {
          h->type = (OldType==PARALLEL?h->type:OldType);
          ParallelSim = TRUE;
        }
      }
      break;
    case 'o':  /* output filename */
      *outptr=fopen(optarg,"w+");
      if(*outptr==NULL) {
        perror("ParseArg : unable to open file");
        error = TRUE;
      } else {
        fname=malloc(strlen(optarg)+1);
        strcpy(fname,optarg);
      }
      break;
    case 'c':                  /* angle through which projections transverse */
      AngleTraversed = atof(optarg);
      AngleTraversed *= TWO_PI;
      if ((errno == ERANGE)||(AngleTraversed<0)) {
        perror("SheppLogan : ParseArg : illegal angle transversed ");
        error = TRUE;
      }
      break;
    case 'v':
      verbose=TRUE;
      break;
    case ':':                                           /* missing parameter */
      perror("ParseArg : missing parameter in option");
      error = TRUE;
      break;
    case '?':                                              /* illegal option */
      perror("ParseArg : illegal option");
      error = TRUE;
      break;
    case -1:                                           /* all options parsed */
      if(ParallelSim) {
        double span=h->dim[ELE]*h->tau[ELE]; /*distance spanned by projection*/
        h->D=PARALLEL_D_FACTOR*span;

        if(h->type==ANGLE)
          h->tau[ELE]=atan(span/h->D)/h->dim[ELE];
      }

      if (h->dim[PROJ]<=0) {
        perror("SheppLogan : number of projections not set");
        error=TRUE;
      }
      if ((fov<=0.0)) {
        perror("SheppLogan : field of view not set");
        error=TRUE;
      }
      if((h->type != PARALLEL) && (h->D<=0.0)) {
        perror("SheppLogan : focal point not set");
        error=TRUE;
      }
      if((h->type != PARALLEL) && ((fov/2.0)>h->D)) {
        perror("SheppLogan : focal point set inside field of view");
        error=TRUE;
      }
      if (h->dim[ELE]<=0) {
        perror("SheppLogan : number of elements per projection not set");
        error=TRUE;
      }
      if (!(h->type)) {
        perror("SheppLogan : CT mode not set");
        error=TRUE;
      }
      if( ((h->type==ANGLE)||(h->type==COLIN)) && (h->D<=0.0) ) {
        perror("SheppLogan : D not set for fan beam data ");
        error=TRUE;
      }
      if ((h->ndim<=0)||(h->type<=IMAGE_TYPE_NOT_SET)||
          (h->type>=MAX_IMAGE_TYPE))
        error=TRUE;

      if(error)
        USAGE(errormsg);

      switch(h->type) {                  /* set h->tau[ELE], element spacing */
      case PARALLEL:
        h->tau[ELE] = fov/h->dim[ELE];
        break;
      case ANGLE:
        h->tau[ELE] = atan((fov/2.0)/h->D)/(h->dim[ELE]/2);
        break;
      case COLIN:
        h->tau[ELE] = ((fov/2.0)/(cos(asin((fov/2.0)/h->D))))/(h->dim[ELE]/2);
        break;
      default:
        perror("ParseArg : illegal CT mode");
        USAGE(errormsg);
      } 

      if(verbose) {  /* print out mode message */
        fprintf(stderr,"SheppLogan : sizeof(float) = %1d\n",
                sizeof(RECON_FLOAT));

        if(fname)
          fprintf(stderr,"SheppLogan : output filename : %s\n",fname);
        else
          fprintf(stderr,"SheppLogan : output to stdout\n");

        fprintf(stderr,"SheppLogan : %d elements per projection\n",
                h->dim[ELE]);
        fprintf(stderr,"SheppLogan : %d projections\n",h->dim[PROJ]);
        fprintf(stderr,"SheppLogan : %f field of view\n",fov);
        fprintf(stderr,"SheppLogan : %f distance between elements\n",h->tau[ELE]);
        fprintf(stderr,"SheppLogan : %f fraction of circle projections transverse\n",AngleTraversed);

        switch(h->type) {
        case PARALLEL:
          fprintf(stderr,"SheppLogan : parallel beam CT data\n");
          break;
        case ANGLE:
          fprintf(stderr,"SheppLogan : equiangular fan beam CT data\n");
          break;
        case COLIN:
          fprintf(stderr,"SheppLogan : collinear fan beam detector\n");
          break;
        default:
          perror("ParseArg : illegal CT mode");
          USAGE(errormsg);
        }
      } /* print out mode message */

                                    /* angle between projections */
      h->tau[PROJ]=AngleTraversed/(h->lim[HI_PROJ]-h->lim[LO_PROJ]+1);

      free(fname);

      break;
    default :
      errno = ILLEGAL_ARGUMENTS;
      USAGE(errormsg);
    }
  } while(opt != -1);

  return(h);
} /* ParseArg() */
